@@ -210,7 +210,7 @@ sub data {
 
 	    # {matchlen} might contain the length of already matched data
 	    # apply only the newly matched to off_not_fwd
-	    $self->{off_not_fwd}[$dir] -= ( $mlen - ($rules->[0]{matchlen}||0) );
+	    $self->{off_not_fwd}[$dir] -= ( $mlen - ($rules->[0]{matchlen}||0));
 
 	    # the rule is definitely done if we reached rxlen
 	    my $rule_done;
@@ -411,13 +411,13 @@ the direction, e.g. 0 for data from client and 1 for data from server
 =item rxlen
 
 the length of data the regular expression might need for the match.
-E.g.  if the regex is C<qr/foo(?=bar)/> 6 bytes are needed for a successful match,
-even if the regex matches only 3 bytes.
+E.g. if the regex is C<qr/foo(?=bar)/> 6 bytes are needed for a successful
+match, even if the regex matches only 3 bytes.
 
 =item rx
 
 the regular expression itself.
-The regex will be applied against the not-yet-forwarded data with an implicit 
+The regex will be applied against the not-yet-forwarded data with an implicit
 C<\A> in front, so look-behind will not work.
 
 =back
